/*
This file is part of Switchback-Generator
    Copyright (C) 2024  Haomiao Wu

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef SWITCHBACK_SHAPE_OPTIMIZER_H
#define SWITCHBACK_SHAPE_OPTIMIZER_H

#include <iostream>
#include <complex>
#define OPTIM_ENABLE_EIGEN_WRAPPERS
#include "ext/optimlib/optim.hpp"
#include "ext/errorFunc/Faddeeva.hh"

#define ZERO_FORCE 0
#define DOUBLE_WITH_FORCE 0

using namespace std;

class SWITCHBACK_SHAPE_OPTIMIZER 
{
public:
	SWITCHBACK_SHAPE_OPTIMIZER(const REAL freq, const REAL nu) {
		hhat = freq;
		Gamma = 1.0/(1.0 + nu);

		k10tilde = (1.0 - hhat * hhat)/(1.0 + (Gamma - 1.0) * hhat * hhat);
    k30tilde = sqrt(1.5 * k10tilde * (1.0 - k10tilde));
		optimParams = VECTOR(11);
		optimParams(0) = -1.351873918189e-01; // _hhat = 0.58
		optimParams(1) = -1.965280005574e-01;
		optimParams(2) = 5.868466583968e-02;
		optimParams(3) = 3.798712165606e-02;
		optimParams(4) = 2.844585460113e-01;
		optimParams(5) = 7.202627236012e-01;
		optimParams(6) = -1.146791902210e-01;
		optimParams(7) = 4.659736734120e-02;
		optimParams(8) = 1.342677684257e-01;
		optimParams(9) = 3.724649711167e-01;
		optimParams(10) = 6.249326874700e-01;
		a11 = -1.351873918189e-01; //0.58 symmetry fminunc
		a12 = -1.965280005574e-01;
		b11 = 5.868466583968e-02;
		b12 = 3.798712165606e-02;
		c11 = 2.844585460113e-01;
		c12 = 7.202627236012e-01;
		a21 = -1.146791902210e-01;
		b21 = 4.659736734120e-02;
		b22 = 1.342677684257e-01;
		c21 = 3.724649711167e-01;
		c22 = 6.249326874700e-01;
		a22 = sqrt(b22)*exp(c22*c22/4.0/b22)*
          (2 * Gamma /sqrt(M_PI)*k30tilde -
           1.0/sqrt(b21)*exp(-c21*c21/4.0/b21)*a21);
		halfSpacing = 0.9 * M_PI;
	}

	~SWITCHBACK_SHAPE_OPTIMIZER(){};

	REAL k1tilde(REAL stilde) const
  {
    return a11 * exp(-b11 * stilde * stilde) * cos(c11 * stilde) +
    a12 * exp(-b12 * stilde * stilde) * cos(c12 * stilde) + k10tilde;
  }

  REAL k2tilde(REAL stilde) const
  {
    return a21 * exp(-b21 * stilde * stilde) * cos(c21 * stilde) + 
           a22 * exp(-b22 * stilde * stilde) * cos(c22 * stilde);
  }

  REAL k3tilde(REAL stilde) const
  {
    double re1, im1, re2, im2, coeff1, coeff2, erf1, erf2, sign1, sign2;
    re1 = sqrt(b21) * stilde;
    im1 = 0.5 * c21 / sqrt(b21);
    complex<double> comp1(re1, im1);
    re2 = sqrt(b22) * stilde;
    im2 = 0.5 * c22 / sqrt(b22);
    complex<double> comp2(re2, im2);
    coeff1 = 0.5 / Gamma * sqrt(M_PI/b21) * exp(- c21*c21/4.0/b21) * a21;
    coeff2 = k30tilde - coeff1;
    erf1 = ((Faddeeva::erf(comp1)).real());
    erf2 = ((Faddeeva::erf(comp2)).real());
    sign1 = erf1 >= 0.0 ? 1.0:-1.0;
    sign2 = erf2 >= 0.0 ? 1.0:-1.0;
    erf1 = abs(erf1) > 1e30?1e30*sign1:erf1;
    erf2 = abs(erf2) > 1e30?1e30*sign2:erf2;
    return -erf1 * coeff1 -erf2 * coeff2;
  }

	// stilde0: starting stilde 
	// positions: a vector of at which turns the switchbacks appear
  REAL k1tildeMulti(REAL stilde, REAL stilde0, VECTOR positions) const
  {
  	VECTOR stildeList = 2 * M_PI * positions/sqrt(k10tilde*k10tilde + k30tilde*k30tilde);
  	stildeList += stilde0 * VECTOR::Ones(positions.size());
  	// cout<<"stildeList: "<<stildeList.transpose()<<endl;
  	// exit(0);
  	REAL k1 = k10tilde;
  	// REAL sign = 1.0;
  	for(int i = 0; i < positions.size(); i++) 
  	{
  		const REAL s = stilde - stildeList[i];
  		k1 += (a11 * exp(-b11 * s * s) * cos(c11 * s) +
    a12 * exp(-b12 * s * s) * cos(c12 * s));
  		// sign = -sign;
  	}
    return k1;
  }

  REAL k2tildeMulti(REAL stilde, REAL stilde0, VECTOR positions) const
  {
  	VECTOR stildeList = 2 * M_PI * positions/sqrt(k10tilde*k10tilde + k30tilde*k30tilde);
  	stildeList += stilde0 * VECTOR::Ones(positions.size());
  	REAL k2 = 0;
  	REAL sign = 1.0;
  	for(int i = 0; i < positions.size(); i++) 
  	{
  		const REAL s = stilde - stildeList[i];
  		k2 += (a21 * exp(-b21 * s * s) * cos(c21 * s) + 
           a22 * exp(-b22 * s * s) * cos(c22 * s))*sign;
      sign = -sign;
  	}
  	return k2;
  }

  REAL k3tildeMulti(REAL stilde, REAL stilde0, VECTOR positions) const
  {
  	double re1, im1, re2, im2, coeff1, coeff2;
  	im1 = 0.5 * c21 / sqrt(b21);
		im2 = 0.5 * c22 / sqrt(b22);
		coeff1 = 0.5 / Gamma * sqrt(M_PI/b21) * exp(- c21*c21/4.0/b21) * a21;
    coeff2 = k30tilde - coeff1;

  	VECTOR stildeList = 2 * M_PI * positions/sqrt(k10tilde*k10tilde + k30tilde*k30tilde);
  	stildeList += stilde0 * VECTOR::Ones(positions.size());
  	REAL k3 = 0;
  	REAL sign = 1.0;
  	for(int i = 0; i < positions.size(); i++) 
  	{
  		const REAL s = stilde - stildeList[i];
  		re1 = sqrt(b21) * s;
	    complex<double> comp1(re1, im1);
	    re2 = sqrt(b22) * s;
	    complex<double> comp2(re2, im2);
  		k3 += sign * (-((Faddeeva::erf(comp1)).real()) * coeff1 -((Faddeeva::erf(comp2)).real()) * coeff2);
  		sign = -sign;
  	}
  	if(sign > 0)
  		k3 += k30tilde;

    return k3;
  }

  REAL k1tildeDouble(REAL stilde) const
  {
    return k10tilde + a11 * exp(-b11 * (stilde + halfSpacing) * (stilde + halfSpacing)) * cos(c11 * (stilde + halfSpacing)) +
									    a12 * exp(-b12 * (stilde + halfSpacing) * (stilde + halfSpacing)) * cos(c12 * (stilde + halfSpacing)) + 
									    a11 * exp(-b11 * (stilde - halfSpacing) * (stilde - halfSpacing)) * cos(c11 * (stilde - halfSpacing)) +
									    a12 * exp(-b12 * (stilde - halfSpacing) * (stilde - halfSpacing)) * cos(c12 * (stilde - halfSpacing));
  }

  REAL k2tildeDouble(REAL stilde) const
  {
    return a21 * exp(-b21 * (stilde + halfSpacing) * (stilde + halfSpacing)) * cos(c21 * (stilde + halfSpacing)) + 
           a22 * exp(-b22 * (stilde + halfSpacing) * (stilde + halfSpacing)) * cos(c22 * (stilde + halfSpacing)) -
           a21 * exp(-b21 * (stilde - halfSpacing) * (stilde - halfSpacing)) * cos(c21 * (stilde - halfSpacing)) - 
           a22 * exp(-b22 * (stilde - halfSpacing) * (stilde - halfSpacing)) * cos(c22 * (stilde - halfSpacing));
  }

  REAL k3tildeDouble(REAL stilde) const
  {
    double re1, im1, re2, im2, coeff1, coeff2, erf1, erf2, sign1, sign2, b21Cut, b22Cut;
    b21Cut = b21 < 1e-18? 1e-18:b21;
    b22Cut = b22 < 1e-18? 1e-18:b22;
    im1 = 0.5 * c21 / sqrt(b21Cut);
    im2 = 0.5 * c22 / sqrt(b22Cut);
    coeff1 = 0.5 / Gamma * sqrt(M_PI/b21Cut) * exp(- c21*c21/4.0/b21Cut) * a21;
    const REAL signCoeff = coeff1 >= 0.0 ? 1.0:-1.0;
    coeff1 = abs(coeff1) > 1e30?1e30*signCoeff:coeff1;
    // coeff2 = k30tilde*0.5 - coeff1;
    coeff2 = k30tilde - coeff1;
    

    REAL k3 = 0;
    re1 = sqrt(b21Cut) * (stilde + halfSpacing);
    re2 = sqrt(b22Cut) * (stilde + halfSpacing);

    complex<double> comp1(re1, im1);
    complex<double> comp2(re2, im2);
    erf1 = ((Faddeeva::erf(comp1)).real());
    erf2 = ((Faddeeva::erf(comp2)).real());
    sign1 = erf1 >= 0.0 ? 1.0:-1.0;
    sign2 = erf2 >= 0.0 ? 1.0:-1.0;
    erf1 = abs(erf1) > 1e30?1e30*sign1:erf1;
    erf2 = abs(erf2) > 1e30?1e30*sign2:erf2;
    k3 = -erf1 * coeff1 -erf2 * coeff2;
    re1 = sqrt(b21Cut) * (stilde - halfSpacing);
    re2 = sqrt(b22Cut) * (stilde - halfSpacing);
    comp1.real(re1);
    comp2.real(re2);
    // if(isnan(k3)){
    // 	cout<<"187: k3 is nan.\n";
    // 	cout<<"a21: "<<a21<<endl;
    // 	cout<<"b21Cut: "<<b21Cut<<endl;
    // 	cout<<"c21: "<<c21<<endl;
    // 	cout<<"re1: "<<re1<<endl;
    // 	cout<<"re2: "<<re2<<endl;
    // 	cout<<"im1: "<<im1<<endl;
    // 	cout<<"im2: "<<im2<<endl;
    // 	cout<<"coeff1: "<<coeff1<<endl;
    // 	cout<<"coeff2: "<<coeff2<<endl;
    // 	cout<<"0.0*((Faddeeva::erf(comp1)).real()): "<<0.0*((Faddeeva::erf(comp1)).real())<<endl;
    // 	cout<<"Faddeeva::erf(comp2): "<<Faddeeva::erf(comp2)<<endl;
    // 	exit(0);
    // }
    erf1 = ((Faddeeva::erf(comp1)).real());
    erf2 = ((Faddeeva::erf(comp2)).real());
    sign1 = erf1 >= 0.0 ? 1.0:-1.0;
    sign2 = erf2 >= 0.0 ? 1.0:-1.0;
    erf1 = abs(erf1) > 1e30?1e30*sign1:erf1;
    erf2 = abs(erf2) > 1e30?1e30*sign2:erf2;
    k3 += erf1 * coeff1 + erf2 * coeff2 + k30tilde;
    // if(isnan(k3)){
    // 	cout<<"192: k3 is nan.\n";
    // 	exit(0);
    // }
    return k3;
  }

  REAL k1tildeGrad(REAL stilde) const
  {
  	return -a11 * exp(-b11 * stilde * stilde) * 
        (2.0 * b11 * stilde * cos(c11 * stilde) + c11 * sin(c11 * stilde))-
        a12 * exp(-b12 * stilde * stilde) * 
        (2.0 * b12 * stilde * cos(c12 * stilde) + c12 * sin(c12 * stilde));
  }

  REAL k1tildeGradDouble(REAL stilde) const
  {
  	const REAL k11 =  -a11 * exp(-b11 * (stilde + halfSpacing) * (stilde + halfSpacing)) * 
        (2.0 * b11 * (stilde + halfSpacing) * cos(c11 * (stilde + halfSpacing)) + c11 * sin(c11 * (stilde + halfSpacing)))-
        a12 * exp(-b12 * (stilde + halfSpacing) * (stilde + halfSpacing)) * 
        (2.0 * b12 * (stilde + halfSpacing) * cos(c12 * (stilde + halfSpacing)) + c12 * sin(c12 * (stilde + halfSpacing)));
    const REAL k12 = -a11 * exp(-b11 * (stilde - halfSpacing) * (stilde - halfSpacing)) * 
        (2.0 * b11 * (stilde - halfSpacing) * cos(c11 * (stilde - halfSpacing)) + c11 * sin(c11 * (stilde - halfSpacing)))-
        a12 * exp(-b12 * (stilde - halfSpacing) * (stilde - halfSpacing)) * 
        (2.0 * b12 * (stilde - halfSpacing) * cos(c12 * (stilde - halfSpacing)) + c12 * sin(c12 * (stilde - halfSpacing)));
    return k11 + k12;
  }

  REAL k1tildeGrad2(REAL stilde) const
  {
  	const REAL kp2 = - a12 * exp(-b12 * stilde * stilde) * 
        ((c12 * c12 + 2.0 * b12 - 4.0 * b12 * b12 * stilde * stilde) * cos(c12 * stilde) -
        4.0 * b12 * c12 * stilde * sin(c12 * stilde));
    const REAL kp1 = - a11 * exp(-b11 * stilde * stilde) * 
        ((c11 * c11 + 2.0 * b11 - 4.0 * b11 * b11 * stilde * stilde) * cos(c11 * stilde) -
        4.0 * b11 * c11 * stilde * sin(c11 * stilde));
    return kp1 + kp2;
  }

  REAL k1tildeGrad3(REAL stilde) const
  {
  	const REAL kp1 = - a11 * exp(-b11 * stilde * stilde) * 
        ((-12.0*b11*b11*stilde-6.0*b11*c11*c11*stilde+8.0*b11*b11*b11 * stilde*stilde*stilde) *
        cos(c11 * stilde) +
        (12.0*b11*b11*c11*stilde*stilde-6.0*b11*c11-c11*c11*c11) * sin(c11 * stilde));
    const REAL kp2 = - a12 * exp(-b12 * stilde * stilde) * 
        ((-12.0*b12*b12*stilde-6.0*b12*c12*c12*stilde+8.0*b12*b12*b12 * stilde*stilde*stilde) *
        cos(c12 * stilde) +
        (12.0*b12*b12*c12*stilde*stilde-6.0*b12*c12-c12*c12*c12) * sin(c12 * stilde));
    return kp1 + kp2;
  }

  REAL k2tildeGrad(REAL stilde) const
  {
  	const REAL kp1 = - a21 * exp(-b21 * stilde * stilde) * 
        (2.0 * b21 * stilde * cos(c21 * stilde) + c21 * sin(c21 * stilde));
    const REAL kp2 = - a22 * exp(-b22 * stilde * stilde) * 
        (2.0 * b22 * stilde * cos(c22 * stilde) + c22 * sin(c22 * stilde));
    return kp1 + kp2;
  }

  REAL k2tildeGradDouble(REAL stilde) const
  {
  	const REAL kp11 = - a21 * exp(-b21 * (stilde + halfSpacing) * (stilde + halfSpacing)) * 
        (2.0 * b21 * (stilde + halfSpacing) * cos(c21 * (stilde + halfSpacing)) + c21 * sin(c21 * (stilde + halfSpacing)));
    const REAL kp21 = - a22 * exp(-b22 * (stilde + halfSpacing) * (stilde + halfSpacing)) * 
        (2.0 * b22 * (stilde + halfSpacing) * cos(c22 * (stilde + halfSpacing)) + c22 * sin(c22 * (stilde + halfSpacing)));
    const REAL kp12 = - a21 * exp(-b21 * (stilde - halfSpacing) * (stilde - halfSpacing)) * 
        (2.0 * b21 * (stilde - halfSpacing) * cos(c21 * (stilde - halfSpacing)) + c21 * sin(c21 * (stilde - halfSpacing)));
    const REAL kp22 = - a22 * exp(-b22 * (stilde - halfSpacing) * (stilde - halfSpacing)) * 
        (2.0 * b22 * (stilde - halfSpacing) * cos(c22 * (stilde - halfSpacing)) + c22 * sin(c22 * (stilde - halfSpacing)));
    return kp11 + kp21 - kp12 - kp22;
  }

  REAL k2tildeGrad2(REAL stilde) const
  {
  	const REAL kp1 = - a21 * exp(-b21 * stilde * stilde) * 
        ((c21 * c21 + 2.0*b21 - 4.0 * b21 * b21 * stilde * stilde) * cos(c21 * stilde)-
        4.0 * b21 * c21 * stilde * sin(c21 * stilde));
    const REAL kp2 = - a22 * exp(-b22 * stilde * stilde) * 
        ((c22 * c22 + 2.0*b22 - 4.0 * b22 * b22 * stilde * stilde) * cos(c22 * stilde)-
        4.0 * b22 * c22 * stilde * sin(c22 * stilde));
    return kp1 + kp2;
  }

  REAL k3tildeGrad(REAL stilde) const
  {
  	return -1.0/Gamma * k2tilde(stilde);
  }

  REAL k3tildeGrad2(REAL stilde) const
  {
  	return -1.0/Gamma * k2tildeGrad(stilde);
  }

  REAL F1Double(REAL stilde) const
  {
  	const REAL k3 = k3tildeDouble(stilde);
  	const REAL k1 = k1tildeDouble(stilde);
#if ZERO_FORCE  	
  	return k3 * k10tilde - k2tildeGradDouble(stilde) + (Gamma - 1.0) * k3 * k1 - Gamma * k30tilde * k1;
#else
  	return k3 - k2tildeGradDouble(stilde) + (Gamma - 1.0) * k3 * k1;
#endif
  }

  REAL F2Double(REAL stilde) const
  {
  	const REAL k2 = k2tildeDouble(stilde);
#if ZERO_FORCE
  	return k1tildeGradDouble(stilde) + (Gamma - 1.0) * k2 * k3tildeDouble(stilde) - Gamma * k30tilde * k2;
#else
  	return k1tildeGradDouble(stilde) + (Gamma - 1.0) * k2 * k3tildeDouble(stilde);
#endif
  }

  REAL F1(REAL stilde) const
  {
  	const REAL k3 = k3tilde(stilde);
  	const REAL k1 = k1tilde(stilde);

#if ZERO_FORCE
  	std::cout<<"old and wrong F1.\n";
  	exit(0);
  	return k3 - k2tildeGrad(stilde) + (Gamma - 1.0) * k3 * k1 - Gamma * k30tilde / k10tilde * k1;
#endif

    return k3 - k2tildeGrad(stilde) + (Gamma - 1.0) * k3 * k1;
  }

  REAL F1Grad(REAL stilde) const
  {
  	const REAL k3 = k3tilde(stilde);
    const REAL k3p = k3tildeGrad(stilde);
    return k3p - k2tildeGrad2(stilde) + (Gamma - 1.0)*(k3p*k1tilde(stilde) + k3*k1tildeGrad(stilde));
  }

  REAL F2(REAL stilde) const
  {
  	const REAL k2 = k2tilde(stilde);
#if ZERO_FORCE
  	return k1tildeGrad(stilde) + (Gamma - 1.0) * k2 * k3tilde(stilde) - Gamma * k30tilde / k10tilde * k2;
#endif  
  	return k1tildeGrad(stilde) + (Gamma - 1.0) * k2 * k3tilde(stilde);
  }

  REAL F2Grad(REAL stilde) const
  {
  	return k1tildeGrad2(stilde) + 
        (Gamma - 1.0) * (k2tildeGrad(stilde) 
        * k3tilde(stilde) +
        k2tilde(stilde) 
        * k3tildeGrad(stilde));
  }

  REAL F2Grad2(REAL stilde) const
  {
  	return k1tildeGrad3(stilde) + 
        (Gamma - 1.0) * (k2tildeGrad2(stilde) 
        * k3tilde(stilde) +
        k2tilde(stilde) 
        * k3tildeGrad2(stilde) + 
        2.0 * k2tildeGrad(stilde) 
        * k3tildeGrad(stilde));
  }

  REAL forceEquationDensity(REAL stilde) const
  {
#if ZERO_FORCE
  	const REAL f1 = F1(stilde);
    const REAL f2 = F2(stilde);
    return f1 * f1 + f2 * f2;
#else
  	const REAL k1 = k1tilde(stilde);
    const REAL k2 = k2tilde(stilde);
    const REAL k3 = k3tilde(stilde);
    const REAL f1p = F1Grad(stilde);
    const REAL f2p = F2Grad(stilde);
  	const REAL k30tilde2 = k30tilde * k30tilde;
  	REAL coeff2 = (1.0/k10tilde-1.0 + Gamma); coeff2 = coeff2 * coeff2;
  	const REAL Fnorm2 = coeff2 * k30tilde2 * (k10tilde * k10tilde + k30tilde2);
    const REAL f1 = F1(stilde);
    const REAL f2 = F2(stilde);
    const REAL f3sq = Fnorm2 - f1*f1 - f2*f2;
    if(f3sq >= 0.0){
    	const REAL f3 = sqrt(f3sq);
	    REAL ff1 = (f2 * k3 - f3 * k2 - f1p); ff1 = ff1 * ff1;
	    REAL ff2 = (f3 * k1 - f1 * k3 - f2p); ff2 = ff2 * ff2;
	    return ff1 + ff2;
	  }
	  else{
	  	const REAL f3 = sqrt(-f3sq);
	    REAL ff1 = (f2 * k3 - f1p); ff1 = ff1 * ff1; ff1 +=  f3*f3 * k2*k2;
	    REAL ff2 = (- f1 * k3 - f2p); ff2 = ff2 * ff2; ff2 += f3*f3 * k1*k1;
	    return ff1 + ff2;
	    // return 1.0;
	  }
#endif
  }

  REAL energyDensity(REAL stilde) const
  {
  	const REAL k1 = k1tilde(stilde);
    const REAL k2 = k2tilde(stilde);
    const REAL k3 = k3tilde(stilde);
    // const REAL k30tilde2 = k30tilde * k30tilde;
    REAL coeff2 = (1.0/k10tilde-1.0 + Gamma); coeff2 = coeff2 * coeff2;
  	// const REAL Fnorm2 = coeff2 * k30tilde2 * (k10tilde * k10tilde + k30tilde2);
    // const REAL f1 = F1(stilde);
    // const REAL f2 = F2(stilde);
    // const REAL f3sq = Fnorm2 - f1*f1 - f2*f2;
    // const REAL f3 = F3(stilde);
    const REAL cubic = 1 + (Gamma - 1) * k1;
    // if(k1 > 1){
    // 	cout<<"k1: "<<k1<<endl;
    // 	exit(0);
    // }
    const REAL elastic = 0.5 * (k1 - 1) * (k1 - 1) + 0.5 * k2 * k2 + 0.5 * Gamma * k3 * k3;
    // if(f3sq >= 0)
    // 	return elastic - sqrt(f3sq);
    // else
    // 	return sqrt(elastic * elastic - f3sq);

    if(k1 > 1)
    	return sqrt(elastic * elastic - (1 - k1)*cubic*cubic*cubic/Gamma/Gamma * hhat * hhat);
    else
    	return elastic - hhat * sqrt((1 - k1)*cubic*cubic*cubic/Gamma/Gamma);

    // return 0.5 * (k1 - 1) * (k1 - 1) + 0.5 * k2 * k2 + 0.5 * Gamma * k3 * k3 - hhat * sqrt((1 - k1)*cubic*cubic*cubic/Gamma/Gamma);
  }

  REAL energyDensityDoubleWithForce(REAL stilde) const
  {
  	const REAL k1 = k1tildeDouble(stilde);
    const REAL k2 = k2tildeDouble(stilde);
    const REAL k3 = k3tildeDouble(stilde);
    const REAL k30tilde2 = k30tilde * k30tilde;
    REAL coeff2 = (1.0/k10tilde-1.0 + Gamma); coeff2 = coeff2 * coeff2;
  	const REAL Fnorm2 = coeff2 * k30tilde2 * (k10tilde * k10tilde + k30tilde2);
    const REAL f1 = F1Double(stilde);
    const REAL f2 = F2Double(stilde);
    const REAL f3sq = Fnorm2 - f1*f1 - f2*f2;
    // const REAL f3 = F3(stilde);
    // const REAL cubic = 1 + (Gamma - 1) * k1;
    const REAL elastic = 0.5 * (k1 - 1) * (k1 - 1) + 0.5 * k2 * k2 + 0.5 * Gamma * k3 * k3;
    if(isnan(elastic)){
      cout<<"elastic is nan.\n";
      cout<<"k1: "<<k1<<endl;
      cout<<"k2: "<<k2<<endl;
      cout<<"k3: "<<k3<<endl;
      exit(0);
    }
    // if(isnan(f3sq)){
    //   cout<<"f3sq is nan.\n";
    //   exit(0);
    // }
    if(f3sq >= 0)
    	return elastic - sqrt(f3sq);
    else
    	return sqrt(elastic * elastic - f3sq);

    // if(k1 > 1)
    // 	return sqrt(elastic * elastic - (1 - k1)*cubic*cubic*cubic/Gamma/Gamma * hhat * hhat);
    // else
    // 	return elastic - hhat * sqrt((1 - k1)*cubic*cubic*cubic/Gamma/Gamma);
  }

  REAL energyDensityDouble(REAL stilde) const
  {
  	const REAL k1 = k1tildeDouble(stilde);
    const REAL k2 = k2tildeDouble(stilde);
    const REAL k3 = k3tildeDouble(stilde);
    const REAL f1 = F1Double(stilde);
    const REAL f2 = F2Double(stilde);
    // const REAL cubic = 1 + (Gamma - 1) * k1;
    if(isnan(f1)){
    	cout<<"f1 is nan.\n";
    	exit(0);
    }
    if(isnan(f2)){
    	cout<<"f2 is nan.\n";
    	exit(0);
    }
    return 0.5 * (k1 - k10tilde) * (k1 - k10tilde) + 0.5 * k2 * k2 + 0.5 * Gamma * (k3 - k30tilde) * (k3 - k30tilde) + abs(f1) + abs(f2);
    // return 0.5 * (k1 - k10tilde) * (k1 - k10tilde) + 0.5 * k2 * k2 + 0.5 * Gamma * (k3 - k30tilde) * (k3 - k30tilde);
  }

  REAL coarseIntegration(REAL range, REAL delta) const
  {
  	const int numSamples = 2 * int(range / delta);
  	REAL stilde = -range + delta * 0.5;
  	REAL integral = 0.0;
  	for(int i = 0; i < numSamples; i++)
  	{
  		integral += delta * energyDensity(stilde);
  		stilde += delta;
  	}
  	if(isnan(integral)){
  		cout<<"integration is nan.\n";
  		exit(0);
  	}

  	return integral;
  }

  REAL maxDensity(REAL range, REAL delta) const
  {
  	const int numSamples = 2 * int(range / delta) + 1;
  	VECTOR samples(numSamples);
  	REAL stilde = -range;
  	for(int i = 0; i < numSamples; i++)
  	{
  		samples[i] = energyDensity(stilde);

  		stilde += delta;
  	}
  	if(samples.hasNaN())
  	{
  		cout<<"max density has nan.\n";
  		exit(0);
  	}
  	return samples.array().maxCoeff();
  }

    REAL maxDensityDoubleWithForce(REAL range, REAL delta) const
  {
    const int numSamples = 2 * int(range / delta) + 1;
    VECTOR samples(numSamples);
    REAL stilde = -range;
    for(int i = 0; i < numSamples; i++)
    {
      samples[i] = energyDensityDoubleWithForce(stilde);
      stilde += delta;
    }
    if(samples.hasNaN())
    {
      cout<<"max density has nan.\n";
      exit(0);
    }
    return samples.array().maxCoeff();
  }


  REAL symmetryFactor(REAL stepSize, REAL sStart) const // 
  {
  	const int numSteps = int(abs(sStart / stepSize));
  	VECTOR3 d3(k30tilde, 0, k10tilde);
    VECTOR3 d2(0, 1, 0);
    VECTOR3 d1(k10tilde, 0, -k30tilde);
    d3 = d3.normalized(); d1 = d1.normalized();
    REAL stilde = sStart;
    VECTOR3 d3Next = d3, d2Next = d2, d1Next = d1;
    for (int i = 0; i < numSteps; i++)
    {
    	stilde += stepSize;
    	const REAL k1 = k1tilde(stilde);
      const REAL k2 = k2tilde(stilde);
      const REAL k3 = k3tilde(stilde);
      // s = s + stepSize;
      const VECTOR3 kappa = k1 * d1Next + k2 * d2Next + k3 * d3Next;
      const REAL kappaNorm = kappa.norm();
      const REAL rotationAgl = kappaNorm * stepSize;
      const Eigen::AngleAxisd rotMat = Eigen::AngleAxisd(rotationAgl, kappa.normalized());
      d1Next = rotMat * d1;
      d2Next = rotMat * d2;
      d3Next = rotMat * d3;
      d1 = d1Next; d2 = d2Next; d3 = d3Next;
    }

    return sqrt(d3(2) * d3(2) + d3(1) * d3(1));
  }

    REAL symmetryFactorDouble(REAL stepSize, REAL sStart) const // 
  {
  	const int numSteps = int((abs(sStart) + halfSpacing * 2.0) / stepSize);
  	const int numStepsMid = int(abs(sStart) / stepSize);
  	VECTOR3 d3(k30tilde, 0, k10tilde);
    VECTOR3 d2(0, 1, 0);
    VECTOR3 d1(k10tilde, 0, -k30tilde);
    d3 = d3.normalized(); d1 = d1.normalized();
    REAL stilde = sStart;
    VECTOR3 d3Next = d3, d2Next = d2, d1Next = d1;
    REAL factor = 0.0;
    for (int i = 0; i < numSteps; i++)
    {
    	stilde += stepSize;
    	const REAL k1 = k1tildeDouble(stilde - halfSpacing);
      const REAL k2 = k2tildeDouble(stilde - halfSpacing);
      const REAL k3 = k3tildeDouble(stilde - halfSpacing);
      // s = s + stepSize;
      const VECTOR3 kappa = k1 * d1Next + k2 * d2Next + k3 * d3Next;
      const REAL kappaNorm = kappa.norm();
      const REAL rotationAgl = kappaNorm * stepSize;
      const Eigen::AngleAxisd rotMat = Eigen::AngleAxisd(rotationAgl, kappa.normalized());
      d1Next = rotMat * d1;
      d2Next = rotMat * d2;
      d3Next = rotMat * d3;
      d1 = d1Next; d2 = d2Next; d3 = d3Next;
      if(numSteps == numStepsMid)
      	factor += sqrt(d3(2) * d3(2) + d3(1) * d3(1));
    }

    factor += sqrt(d3(2) * d3(2) + d3(1) * d3(1));
    return factor;
  }


    REAL symmetryFactorDouble1(REAL stepSize, REAL sStart) const // 
  {
  	const int numSteps = int(abs(sStart) * 2.0  / stepSize);
  	VECTOR3 d3(k30tilde, 0, k10tilde);
    VECTOR3 d2(0, 1, 0);
    VECTOR3 d1(k10tilde, 0, -k30tilde);
    d3 = d3.normalized(); d1 = d1.normalized();
    REAL stilde = sStart;
    VECTOR3 d3Next = d3, d2Next = d2, d1Next = d1;
    VECTOR3 kappa;
    REAL k1, k2, k3;
    REAL factor = 0.0;
    for (int i = 0; i < numSteps; i++)
    {
    	stilde += stepSize;
    	k1 = k1tildeDouble(stilde - halfSpacing);
      k2 = k2tildeDouble(stilde - halfSpacing);
      k3 = k3tildeDouble(stilde - halfSpacing);
      // stilde += stepSize;
      // s = s + stepSize;
      kappa = k1 * d1Next + k2 * d2Next + k3 * d3Next;
      const REAL kappaNorm = kappa.norm();
      const REAL rotationAgl = kappaNorm * stepSize;
      const Eigen::AngleAxisd rotMat = Eigen::AngleAxisd(rotationAgl, kappa.normalized());
      d1Next = rotMat * d1;
      d2Next = rotMat * d2;
      d3Next = rotMat * d3;
      d1 = d1Next; d2 = d2Next; d3 = d3Next;
    }
    kappa = k1 * d1Next + k2 * d2Next + k3 * d3Next;
    kappa = kappa.normalized();
    factor += 1.0 - kappa(0);
    // factor += sqrt(kappa(2) * kappa(2) + kappa(1) * kappa(1));
    return factor;
  }

  REAL totalEnergy(const VECTOR& input) 
  {
  	a11 = input(0);
  	a12 = input(1);
  	b11 = abs(input(2));
  	b12 = abs(input(3));
  	c11 = input(4);
  	c12 = input(5);
  	a21 = input(6);
  	b21 = abs(input(7));
  	b22 = abs(input(8));
  	c21 = input(9);
  	c22 = input(10);
  	a22 = sqrt(b22)*exp(c22*c22/4.0/b22)*
          (2 * Gamma /sqrt(M_PI)*k30tilde -
           1.0/sqrt(b21)*exp(-c21*c21/4.0/b21)*a21);
    const REAL forceDensity = maxDensity(20.0, 1.0e-1);
    const REAL sym = symmetryFactor(4.0e-2, -2.0 * M_PI * 3.0);
    return forceDensity + 1.0 * sym;
  }

  REAL totalEnergyDouble(const VECTOR& input) 
  {
  	a11 = input(0);
  	a12 = input(1);
  	b11 = abs(input(2));
  	b12 = abs(input(3));
  	c11 = input(4);
  	c12 = input(5);
  	a21 = input(6);
  	b21 = abs(input(7));
  	b22 = abs(input(8));
  	c21 = input(9);
  	c22 = input(10);
  	// halfSpacing = input(11);
  	a22 = sqrt(b22)*exp(c22*c22/4.0/b22)*
          (2 * Gamma /sqrt(M_PI)*k30tilde -
           1.0/sqrt(b21)*exp(-c21*c21/4.0/b21)*a21);
    // a22 = sqrt(b22)*exp(c22*c22/4.0/b22)*
    // (2 * Gamma /sqrt(M_PI)*k30tilde*0.5 -
    //  1.0/sqrt(b21)*exp(-c21*c21/4.0/b21)*a21);
    const REAL forceDensity = maxDensity(20.0, 1.0e-1);
    const REAL sym1 = symmetryFactorDouble1(4.0e-2, -2.0 * M_PI * 3.0);
    const REAL sym = symmetryFactorDouble(4.0e-2, -2.0 * M_PI * 3.0);
    return forceDensity + 0.8 * sym + 0.8 * sym1 + 0.10 * (abs(k3tildeDouble(0)) + abs(k1tildeDouble(0)));
  }

  REAL totalEnergyDoubleWithForce(const VECTOR& input) 
  {
  	a11 = input(0);
  	a12 = input(1);
  	b11 = abs(input(2));
  	b12 = abs(input(3));
  	c11 = input(4);
  	c12 = input(5);
  	a21 = input(6);
  	b21 = abs(input(7));
  	b22 = abs(input(8));
  	c21 = input(9);
  	c22 = input(10);

    // b11 = b11 < 1e-8? 1e-8:b11;
    // b12 = b12 < 1e-8? 1e-8:b12;
    // b21 = b21 < 1e-8? 1e-8:b21;
    // b22 = b22 < 1e-8? 1e-8:b22;
  	// halfSpacing = abs(input(11));
  	a22 = sqrt(b22)*exp(c22*c22/4.0/b22)*
          (2 * Gamma /sqrt(M_PI)*k30tilde -
           1.0/sqrt(b21)*exp(-c21*c21/4.0/b21)*a21);
    // if(isnan(a22)){
    //   cout<<"a22 is nan.\n";
    //   cout<<"params: "<<input.transpose()<<endl;
    //   cout<<"b11: "<<b11<<endl;
    //   cout<<"c22: "<<c22<<endl;
    //   cout<<"(c22*c22/4.0/b22): "<<(c22*c22/4.0/b22)<<endl;
    //   exit(0);
    // }
    const REAL forceDensity = maxDensityDoubleWithForce(20.0, 1.0e-1);
    const REAL sym1 = symmetryFactorDouble1(4.0e-2, -2.0 * M_PI * 3.0);
    const REAL sym = symmetryFactorDouble(4.0e-2, -2.0 * M_PI * 3.0);
    // return forceDensity + 10.0 * sym1 + 0.5 * sym;  // hhat = 0.3
    return forceDensity + 10.0 * sym1 + 0.3 * sym;  // hhat = 0.4
  }

  // void runOptimization(){
  // 	bool success = optim::de(optimParams, &(SWITCHBACK_SHAPE_OPTIMIZER::totalEnergy), nullptr);
  // 	if(success) {
  // 		cout<<"optimization successed.\n";
  // 	}
  // 	else {
  // 		cout<<"optimization failed.\n";
  // 	}
  // }

	REAL a11, b11, c11, a21, b21, c21;
  REAL a12, b12, c12, a22, b22, c22;
  REAL hhat, k30tilde, k10tilde, Gamma, halfSpacing; // Gamma = 1/(1+nu)
  VECTOR optimParams;
};



#endif

