/*
This file is part of Switchback-Generator
    Copyright (C) 2024  Haomiao Wu

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
#include <iostream>
#include <fstream>
#include <complex>
#include <string>
#include "ext/errorFunc/Faddeeva.hh"
#include <Eigen/Dense>
#include <Eigen/Sparse>

typedef double REAL;

typedef Eigen::Matrix<REAL, 2,  2>  MATRIX2;
typedef Eigen::Matrix<REAL, 3,  2>  MATRIX3x2;
typedef Eigen::Matrix<REAL, 3,  3>  MATRIX3;
typedef Eigen::Matrix<REAL, 3,  6>  MATRIX3x6;
typedef Eigen::Matrix<REAL, 3,  12> MATRIX3x12;
typedef Eigen::Matrix<REAL, 4,  4>  MATRIX4;
typedef Eigen::Matrix<REAL, 6,  6>  MATRIX6;
typedef Eigen::Matrix<REAL, 6,  4>  MATRIX6x4;
typedef Eigen::Matrix<REAL, 6,  9>  MATRIX6x9;
typedef Eigen::Matrix<REAL, 6,  11>  MATRIX6x11;
typedef Eigen::Matrix<REAL, 6,  12> MATRIX6x12;
typedef Eigen::Matrix<REAL, 9,  9>  MATRIX9;
typedef Eigen::Matrix<REAL, 9,  4>  MATRIX9x4;
typedef Eigen::Matrix<REAL, 9,  12> MATRIX9x12;
typedef Eigen::Matrix<REAL, 11,  2>  MATRIX11x2;
typedef Eigen::Matrix<REAL, 11, 11>  MATRIX11;
typedef Eigen::Matrix<REAL, 12, 12> MATRIX12;
typedef Eigen::Matrix<REAL, 2,  1>  VECTOR2;
typedef Eigen::Matrix<REAL, 3,  1>  VECTOR3;
typedef Eigen::Matrix<REAL, 4,  1>  VECTOR4;
typedef Eigen::Matrix<REAL, 6,  1>  VECTOR6;
typedef Eigen::Matrix<REAL, 9,  1>  VECTOR9;
typedef Eigen::Matrix<REAL, 11,  1>  VECTOR11;
typedef Eigen::Matrix<REAL, 12, 1>  VECTOR12;

typedef Eigen::Matrix<int, 2, 1> VECTOR2I;
typedef Eigen::Matrix<int, 3, 1> VECTOR3I;
typedef Eigen::Matrix<int, 4, 1> VECTOR4I;
typedef Eigen::Matrix<int, 5, 1> VECTOR5I;
typedef Eigen::Matrix<int, Eigen::Dynamic, 1> VECTORI;

typedef Eigen::Matrix<REAL, Eigen::Dynamic, 1> VECTOR;
typedef Eigen::Matrix<REAL, Eigen::Dynamic, Eigen::Dynamic> MATRIX;
typedef Eigen::SparseMatrix<REAL, Eigen::RowMajor> SPARSE_MATRIX;

#include "SWITCHBACK_SHAPE_OPTIMIZER.h"

using namespace std;

REAL switchbackTotalEnergyHandle(const VECTOR& vals_inp, VECTOR* grad_out, void* opt_data)
{
	SWITCHBACK_SHAPE_OPTIMIZER* shape = static_cast<SWITCHBACK_SHAPE_OPTIMIZER*>(opt_data);
	return shape->totalEnergy(vals_inp);
}

REAL switchbackTotalEnergyHandleDouble(const VECTOR& vals_inp, VECTOR* grad_out, void* opt_data)
{
	SWITCHBACK_SHAPE_OPTIMIZER* shape = static_cast<SWITCHBACK_SHAPE_OPTIMIZER*>(opt_data);
	return shape->totalEnergyDouble(vals_inp);
}

REAL switchbackTotalEnergyHandleDoubleWithForce(const VECTOR& vals_inp, VECTOR* grad_out, void* opt_data)
{
	SWITCHBACK_SHAPE_OPTIMIZER* shape = static_cast<SWITCHBACK_SHAPE_OPTIMIZER*>(opt_data);
	return shape->totalEnergyDoubleWithForce(vals_inp);
}

bool writeOBJFile(const string& fileName, const vector<VECTOR3>& vertices){
	std::ofstream outFile(fileName);
  if (!outFile.is_open()) {
    std::cerr << "IOError: writeOBJFile() could not open "
              << fileName << std::endl;
    return false;
  }

  for(int i = 0; i < vertices.size(); i++){
  	const VECTOR3& vertex = vertices[i];
    outFile << "v "<< vertex(0) << " "<<vertex(1)<<" "<<vertex(2)<<"\n";
  }

  outFile<<"l";
  for(int i = 0; i < vertices.size(); i++) {
  	outFile << " " << (i + 1);
  }
  outFile<<"\n";
  outFile.close();

  return true;
}

int main(int argc, char** argv) {
	// input parameters
	REAL _hhat = 0.4;
	int isDouble = 0;
	if(argc == 2) {
		_hhat = (REAL)stod(argv[1]);
	}
	if(argc == 3) {
		_hhat = (REAL)stod(argv[1]);
		isDouble = stoi(argv[2]);
	}

	string fileName = "./switchback.obj";

	// material parameters
  const REAL _nu = 0.5; // Poisson's ratio
	const REAL _Gamma = 1.0/(1.0 + _nu);

	// geometry parameters
	const REAL _turns = 10.0;
  const REAL _phi = 2.0 * M_PI * _turns; 
  const REAL _R = 0.05;
  const REAL _numVert = 1000 + 1;
  const REAL _sSubStep = 100;
	const REAL _k0 = 1.0/_R;
  const REAL _k10tilde = (1.0 - _hhat * _hhat)/(1.0 + (_Gamma - 1.0) * _hhat * _hhat);
  const REAL _k30tilde = sqrt(1.0/_Gamma * _k10tilde * (1.0 - _k10tilde));
  const REAL _k30 = _k0 * _k30tilde;
  const REAL _k10 = _k0 * _k10tilde;
  const VECTOR positions {{4.5}}; // switchback positions (in turns)

  // optimization
  SWITCHBACK_SHAPE_OPTIMIZER optimizer(_hhat, _nu);
  VECTOR params = optimizer.optimParams;
	cout<<"original optimizer.totalEnergyDouble(params): "<<optimizer.totalEnergyDoubleWithForce(params)<<endl;
	// cout<<"original optimizer48.totalEnergy(params): "<<optimizer48.totalEnergy(params)<<endl;
	optim::algo_settings_t settings;
	settings.iter_max = 1e6;
	settings.rel_objfn_change_tol = 1e-8;
	settings.rel_sol_change_tol = 1e-10;
	settings.print_level = 0;
	(settings.nm_settings).par_alpha = 1.0;
	(settings.nm_settings).par_beta = 0.5;
	(settings.nm_settings).par_gamma = 1.5;
	(settings.nm_settings).par_delta = 0.5;
	bool success = true;
	if(isDouble){
		params.resize(12);
		params(0) = 0.0027172;
		params(1) = -0.117061;
		params(2) = -1.53057e-08;
		params(3) = 0.0162536;
		params(4) = 0.475034;
		params(5) = 0.915471;
		params(6) = -0.160969;
		params(7) = 0.0246021;
		params(8) = 0.078499;
		params(9) = 0.524765;
		params(10) = 0.710822;
		params(11) = optimizer.halfSpacing;
		success = optim::nm(params, switchbackTotalEnergyHandleDoubleWithForce, &optimizer, settings);
	}
	else{
		success = optim::nm(params, switchbackTotalEnergyHandle, &optimizer, settings);
	}	

	if(success){
	cout<<"a11 = "<<params(0)<<endl<<"a12 = "<<params(1)<<endl<<"b11 = "<<params(2)<<endl<<"b12 = "<<params(3)<<endl<<"c11 = "<<params(4)<<endl<<"c12 = "<<params(5)<<endl
	<<"a21 = "<<params(6)<<endl<<"b21 = "<<params(7)<<endl<<"b22 = "<<params(8)<<endl<<"c21 = "<<params(9)<<endl<<"c22 = "<<params(10)<<endl;
	if(params.size() == 12)
	  cout<<"halfSpacing = "<<params(1)<<endl;
	cout<<"settings.opt_iter: "<<settings.opt_iter<<endl;
	cout<<"settings.opt_error_value: "<<settings.opt_error_value<<endl;
	cout<<"settings.opt_fn_value: "<<settings.opt_fn_value<<endl;

	}
	else{
	  cout<<"optimization failed."<<endl;
	  exit(0);
	}


	// integrate to generate shape
	const REAL klambda = sqrt(_k10 * _k10 + _k30 * _k30);
  const REAL segSize = _phi / (REAL)(_numVert - 1)/_k0;
  const REAL stepSize = segSize / (REAL)_sSubStep;

  // director initial values
  VECTOR3 d1(_k10tilde, 0.0, -_k30tilde);
  VECTOR3 d2(0.0, 1.0, 0.0);
  VECTOR3 d3(_k30tilde, 0.0, _k10tilde);
  VECTOR3 d1New, d1Next;
  VECTOR3 d2New, d2Next;
  VECTOR3 d3New, d3Next;
  VECTOR3 kappa;
  d1 = d1.normalized();
  d2 = d2.normalized();
  d3 = d3.normalized();

  vector<VECTOR3> _restVertices;

  VECTOR3 pos(0.0, 0.0, 0.0);
  REAL s = - _phi * 0.5 / klambda;
  REAL stilde = s * _k0;
  const REAL stilde0 = stilde;
  REAL k1 = _k10, k2 = 0.0, k3 = _k30;
  _restVertices.push_back(pos);


  // const int maxIter = 20;
    // const REAL tol = stepSize * 1e-12;
    for(int vId = 1; vId < _numVert; vId ++) {
      for(int i = 0; i < _sSubStep; i++){
        // get d(s + h) (forward integration)
        d1New = d1; d2New = d2; d3New = d3;
				if(isDouble){
	        k1 = _k0 * optimizer.k1tildeDouble(stilde);
	        k2 = _k0 * optimizer.k2tildeDouble(stilde);
	        k3 = _k0 * optimizer.k3tildeDouble(stilde);
	      }
	      else{
	        k1 = _k0 * optimizer.k1tildeMulti(stilde, stilde0, positions);
	        k2 = _k0 * optimizer.k2tildeMulti(stilde, stilde0, positions);
	        k3 = _k0 * optimizer.k3tildeMulti(stilde, stilde0, positions);
	      }

        //////// rotation matrix integration ///////
        kappa = k1 * d1 + k2 * d2 + k3 * d3;
        const REAL rotationAgl = kappa.norm() * stepSize;
        s += stepSize; stilde = s * _k0;
        const Eigen::AngleAxisd rotMat = Eigen::AngleAxisd(rotationAgl, kappa.normalized());
        d1Next = rotMat * d1;
        d2Next = rotMat * d2;
        d3Next = rotMat * d3;
        ////////////////////////////////////////////

        d1 = d1Next.normalized(); d2 = d2Next.normalized(); d3 = d3Next.normalized();

        // orthogonalize
        d1 = (d1 - d2 * (d1.dot(d2))).normalized();
        d3 = (d3 - d2 * (d3.dot(d2))).normalized();
        d3 = (d3 - d1 * (d3.dot(d1))).normalized();

        // x(s + h) = x(s) + h * d3(s + h)
        pos = pos + stepSize * d3;
      }
      _restVertices.push_back(pos);
    }

    cout<<"Total vertices: "<<_restVertices.size()<<endl;
    if(writeOBJFile(fileName, _restVertices)){
    	cout<<"Wrote to file "<<fileName<<endl;
    }



	return 0;
}
